/* am930llc.c: Handles the Linux network interface
*	--------------------------------------------------------------------
*
*   Linux WLAN 
*
*   The contents of this file are subject to the Mozilla Public
*   License Version 1.0 (the "License"); you may not use this file
*   except in compliance with the License. You may obtain a copy of
*   the License at http://www.mozilla.org/MPL/
*
*   Software distributed under the License is distributed on an "AS
*   IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
*   implied. See the License for the specific language governing
*   rights and limitations under the License.
*
*   The initial developer of the original code is Mark S. Mathews
*   <mark@absoval.com>.  Portions created by Mark S. Mathews
*   are Copyright (C) 1998 AbsoluteValue Software, Inc.  All Rights Reserved.
*   
*	--------------------------------------------------------------------
*
*	The author may be reached as mark@absoval.com, or C/O AbsoluteValue
*	Software Inc., P.O. Box 941149, Maitland, FL, 32794-1149
*
*	Thanks to David Hinds, Donald Becker, and the rest of the Linux
*	developers worldwide for making all of this possible.
*
*	--------------------------------------------------------------------
*
*	Seperate file isolating the functions that handle the Linux network
*	interface. LLC may not be the appropriate name, but this code _does_
*	sit on top of the MAC...
*
*/

/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

/* PCMCIA headers generated during PCMCIA package installation */
#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

/* Module related headers, non-module drivers should not include */
#include <linux/version.h>

/* Standard driver includes */
#include <assert.h>
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <linux/errno.h>

/* Ethernet and network includes */
#include <linux/if_ether.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/ioport.h>

/* Card and Driver services includes */
#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>
#include <pcmcia/cisreg.h>

/* Local Includes */
#include <wlan/version.h>
#include <wlan/am930const.h>
#include <wlan/am930mib.h>
#include <wlan/p80211.h>
#include <wlan/wlan_ioctl.h>
#include "am930debug.h"
#include "am930mac.h"
#include "am930hw.h"
#include "am930mgr.h"
#include "am930llc.h"

/*================================================================*/
/* Local Macros  */

#define FILEBIT	DBLLC

/*================================================================*/
/* Local Types  */

typedef struct enet_statistics	enet_stats_t;

typedef struct devpriv
{
	enet_stats_t	stats;
	am930llc_t*		llc;

} devpriv_t;

typedef int (*do_ioctl_t)(struct device *dev, struct ifreq *ifr, int cmd);

/*================================================================*/
/* Local Functions  */

static int am930llc_devinit(device_t *dev);
static int am930llc_devopen(device_t *dev);
static int am930llc_devstop(device_t *dev);
static int am930llc_devhard_start_xmit( struct sk_buff *skb, device_t *dev);
static enet_stats_t* am930llc_devgetstats(device_t *dev);
static void am930llc_devset_multicast_list(device_t *dev);
int am930llc_devdo_ioctl(device_t *dev, wlan_req_t *req, int cmd);

void am930llc_dbprt_frame( UINT8* frm, UINT32 len, int isrx );

/*================================================================*/
/* Static variables  */


/*================================================================*/

/*----------------------------------------------------------------
*	am930llc_construct
*
*	returns: addr. of the lnd object if successful, NULL otherwise
----------------------------------------------------------------*/
am930llc_t *am930llc_construct(am930mac_t *mac, am930mgr_t *mgr)
{
	am930llc_t *llc;

	DBFENTER;

	llc = kmalloc( sizeof(am930llc_t), GFP_KERNEL);
	if ( llc != NULL )
	{
		llc->mac = mac;
		llc->mgr = mgr;

		/* lets begin the initialization of the linux device */
		llc->dev = kmalloc( sizeof(device_t), GFP_KERNEL);
		memset( llc->dev, 0, sizeof(device_t));

		llc->dev->priv = kmalloc( sizeof(devpriv_t), GFP_KERNEL);
		memset( llc->dev->priv, 0, sizeof(devpriv_t));

		/* the dev name field will point to the same memory used by the 
			node_t field of the pcmcia di object */
		llc->dev->name = llc->mac->di->dev->dev_name;

		/* set the private data to point back at the llc */
		V2P(llc->dev->priv)->llc = llc;

		/* set the hardware address */
		am930mgr_mibgetitem( llc->mgr, ADDR_MAC_ADDR, llc->dev->dev_addr, 6);

		/* set the method pointers */
		llc->dev->init					= am930llc_devinit;
		llc->dev->open					= am930llc_devopen;
		llc->dev->stop					= am930llc_devstop;
		llc->dev->hard_start_xmit 		= am930llc_devhard_start_xmit;
		llc->dev->get_stats				= am930llc_devgetstats;
		llc->dev->set_multicast_list	= am930llc_devset_multicast_list;
		llc->dev->do_ioctl				= (do_ioctl_t)am930llc_devdo_ioctl;

		ether_setup( llc->dev );

		if ( register_netdev(llc->dev) != 0 )
		{
			DBPRT( DBINIT, "register_netdev failed!\n");
			kfree( llc->dev );
			llc->dev = NULL;
			am930llc_destruct( llc );
			llc = NULL;
		}
		DBPRT( DBINIT, "Device %s registered\n", llc->dev->name);
	}

	DBFEXIT

    return llc;
} 


/*----------------------------------------------------------------
*	am930llc_destruct
*
*	returns: nothing
----------------------------------------------------------------*/
void am930llc_destruct( am930llc_t *llc)
{
	DBFENTER;

	if ( llc->dev != NULL )
	{
		llc->dev->start = 0;
		unregister_netdev( llc->dev );
		kfree(llc->dev);
		llc->dev = NULL;
	}

	kfree_s( llc, sizeof(am930llc_t));

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930llc_devinit
*	init method for the linux net device. Called by 
*	register_netdevice. For this driver, it doesn't have to do
*	anything.
*
*	returns: zero
----------------------------------------------------------------*/
int am930llc_devinit(device_t *dev)
{
	int result = 0;
	DBFENTER;

	DBFEXIT;
	return result;
}


/*----------------------------------------------------------------
*	am930llc_devopen
*	open method for the linux net device. Called when ifconfig
*	is used to set up the interface.
*
*	returns: zero
----------------------------------------------------------------*/
int am930llc_devopen(device_t *dev)
{
	int result = 0;
	DBFENTER;

	/* set the flags in the device object */
	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;

	MOD_INC_USE_COUNT;

	DBFEXIT;
	return result;
}


/*----------------------------------------------------------------
*	am930llc_devstop
*	stop method for the linux net device. Called  when ifconfig
*	is used to shut down an interface.
*
*	returns: zero
----------------------------------------------------------------*/
int am930llc_devstop(device_t *dev)
{
	int result = 0;

	DBFENTER;

	/* set the flags in the device object */
	dev->start = 0;
	dev->tbusy = 1;

	MOD_DEC_USE_COUNT;

	DBFEXIT;
	return result;
}


/*----------------------------------------------------------------
*	am930llc_devhard_start_xmit
*	hard_start_xmint method for the linux net device. Called by 
*	the higher level protocol code when it has a packet to xmit.
*
*	returns: zero
----------------------------------------------------------------*/
int am930llc_devhard_start_xmit( struct sk_buff *skb, device_t *dev)
{
	int result = 0;

	DBFENTER;

	if ( dev->start == 1 )
	{
		/* If some higher layer thinks we've missed a tx-done, we are passed
			NULL. Caution: dev_tint handles the cli/sti ..
		*/
#if (LINUX_VERSION_CODE < VERSION(2,1,25))
		if ( skb == NULL )
		{
			dev->tbusy = 0;
			dev_tint(dev);
			return 0;
		}
#endif

		if ( dev->tbusy ) /* shouldn't ever happen */
		{
			/* printk(KERN_DEBUG"xmit called when tbusy set!\n"); */

			/* need to add a reset capability here. see skeleton.c */

			result = -EBUSY;
		}
		else
		{
#ifdef DEBUG_TXHDR_ETH
am930llc_dbprt_frame( skb->data, skb->len, 0 );
#endif
			dev->trans_start = jiffies;

			result = am930mac_txdata( V2P(dev->priv)->llc->mac, 
							 skb->data,
							 skb->data + 6,
							 skb->data, skb->len);

			if ( result )
			{
				dev->tbusy = 1;
			}
		}

		if ( !result )
#if (LINUX_VERSION_CODE < VERSION(2,2,0))
			dev_kfree_skb( skb, FREE_WRITE);
#else
		  dev_kfree_skb( skb );
#endif

	}

	DBFEXIT;
	return result;
}


/*----------------------------------------------------------------
*	am930llc_rxframe
*	Event method called from mac when a frame has been received.
*
*	returns: zero
----------------------------------------------------------------*/
void am930llc_rxframe( am930llc_t *llc, UINT8 *frm, UINT32 len)
{
	struct sk_buff	*skb;

	DBFENTER;

	llc->dev->interrupt = 1;

#ifdef DEBUG_RXHDR_ETH
am930llc_dbprt_frame( frm, len, 1 );
#endif

	if ( llc->dev->start )
	{
		skb = dev_alloc_skb(len + 2); /* I do +2 because it's what 8390 does */
		if ( skb == NULL )
		{
			DBPRT( DBRXPATH, "Memory low, frame dropped\n");
		}
		else
		{
			llc->dev->last_rx = jiffies;

			skb->len = len;
			skb->dev = llc->dev;

			/* The following is from 8390.c */
			skb_reserve(skb, 2);	/* IP headers on 16 byte boundaries */
			skb_put( skb, len);		/* make room */
			memcpy( skb->data, frm, len);
			skb->protocol = eth_type_trans( skb, llc->dev);

/*
#define BLOCK_INPUT(buf, len) { memcpy( (buf), (frm), (len)); }

			GET_PACKET( llc->dev, skb, len);
*/
			V2P(llc->dev->priv)->stats.rx_packets++;

			netif_rx(skb);
			mark_bh(NET_BH);
		}
	}

	llc->dev->interrupt = 0;

	DBFEXIT;
	return;
}

/*----------------------------------------------------------------
*	am930llc_rxframe_err
*
*	returns: zero
----------------------------------------------------------------*/
void am930llc_rxframe_err( am930llc_t *llc)
{
	DBFENTER;

	V2P(llc->dev->priv)->stats.rx_errors++;

	DBFEXIT;
	return;
}

/*----------------------------------------------------------------
*	am930llc_devgetstats
*	getstats method for the linux net device. Called as a result
*	of a number of user mode utilities used to check the rx/tx
*	statistics of an interface.
*
*	returns: zero
----------------------------------------------------------------*/
enet_stats_t* am930llc_devgetstats(device_t *dev)
{
	DBFENTER;
	DBFEXIT;
	return &(V2P(dev->priv)->stats);
}


/*----------------------------------------------------------------
*	am930llc_devset_multicast_list
*	set_multicast_list method for the linux net device. 
*
*	returns: nothing
----------------------------------------------------------------*/
#define HW_MAX_ADDRS 4
void am930llc_devset_multicast_list(device_t *dev)
{
	DBFENTER;
	
	if ( dev->flags & IFF_PROMISC )
	{
		/* Enable Promiscuous Mode */
		UINT8 one = 1;
		am930mgr_mibsetitem(V2P(dev->priv)->llc->mgr, 
			MAC_PROMISC_EN, &one, sizeof(one));

	}
	else if ( dev->flags & IFF_ALLMULTI || (dev->mc_count > HW_MAX_ADDRS) )
	{
		/* Enable the hw "rcv all multicast" mode, either because of 
		   the IFF_ALLMULTI switch or because we have more multicast
		   hw addresses than the hw filter can hold (1.0 fw has four).
		*/
		printk(KERN_DEBUG "IFF_ALLMULTI currently unsupported.\n");
	}
	else if ( dev->mc_count )
	{
		/* Load the contents of the mc_list from the dev structure into
		   the filter list in the hardware (effectively reset it from
		   scratch).
		*/
		printk(KERN_DEBUG "Multicast filters currently unsupported.\n");
	}
	else
	{
		/* Disable promiscuous, disable allmulti, and clear the multicast
		   filter list
		*/
		UINT8 zero = 0;

		am930mgr_mibsetitem(V2P(dev->priv)->llc->mgr, 
			MAC_PROMISC_EN, &zero, sizeof(zero));

	}

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930llc_devdo_ioctl
*	do_ioctl method for the linux net device. Private ioctl handler
*	for the net device. Used here to pass commands to the mgr
*	object.
*	TODO: Define the interface between the mgr object and user
*			mode utilities. 
*
*	returns: zero
----------------------------------------------------------------*/
int am930llc_devdo_ioctl(device_t *dev, wlan_req_t *req, int cmd)
{
	int 	result = 0;
	DBFENTER;

	/* Test the memory passed */
	if ( req->len > 0 )
		result = verify_area( VERIFY_WRITE, req->data, req->len);

	if ( result == 0 )
	{
		switch( cmd )
		{
			case WLAN_TEST:
				req->result = 0xf0f0;
				break;

			/* start mod: tld */
		        case WLAN_TLD:
			  req->result = am930hw_rssi_tld(30);
			  break;
		        /* end mod: tld */
		    
			case WLAN_SCAN:
			{
				wlan_scan_t cmd;
				copy_from_user( &cmd, req->data, sizeof(cmd));
				req->result = am930mgr_scan( V2P(dev->priv)->llc->mgr,
										cmd.scantype,	/*active/passive*/
										cmd.bsstype,	/*ind./infra./both*/
										cmd.bssid,		/*bcast/specific*/
										cmd.startch,	
										cmd.endch,
										cmd.timech );	/*time per ch.*/
				break;
			}
			case WLAN_NETLIST_LEN:
			{	
				wlan_netlist_len_t cmd;
				knownbss_t		*curr;
				int				i;
				am930mgr_t*		mgr = V2P(dev->priv)->llc->mgr;
				i = 0;
				for( curr = mgr->bsslist; curr != NULL; curr = curr->next)
					i++;

				cmd.nitems = i;
				copy_to_user( req->data, &cmd, sizeof(cmd));
				req->result = 0;
				break;
			}
			case WLAN_NETLIST:
			{
				wlan_netlist_t 	*cmd;
				wlan_netlist_t 	tmp;
				knownbss_t		*curr;
				int				i;
				am930mgr_t*		mgr = V2P(dev->priv)->llc->mgr;

				copy_from_user( &tmp, req->data, sizeof(tmp));
				cmd = kmalloc(sizeof(wlan_netlist_t) + (tmp.nitems * sizeof(netitem_t)), GFP_ATOMIC);
				cmd->nitems = tmp.nitems;
				i = 0;
				for( curr = mgr->bsslist; curr != NULL; curr = curr->next)
				{
					cmd->netlist[i].channel = curr->channel;
					cmd->netlist[i].bcn_int = curr->bcn_int;
					memcpy(cmd->netlist[i].bssid, curr->bssid, 6);
					strcpy(cmd->netlist[i].ssid, curr->ssid);
					i++;
				}
				copy_to_user( req->data, cmd, 
					sizeof(wlan_netlist_t) + (cmd->nitems * sizeof(netitem_t)));
				break;
			}
			case WLAN_BSSCREATE:
			{
				wlan_bsscreate_t cmd;
				copy_from_user( &cmd, req->data, sizeof(cmd));
				
				req->result = am930mgr_createbss( V2P(dev->priv)->llc->mgr, 
										cmd.channel, 
										cmd.ssid,
										cmd.beacon_int, 
										cmd.atim_win);
				break;
			}
			case WLAN_BSSJOIN:
			{
				wlan_bssjoin_t	cmd;
				copy_from_user( &cmd, req->data, sizeof(cmd));
				req->result = am930mgr_joinbss( V2P(dev->priv)->llc->mgr,
								cmd.bssid );
				break;
			}
			case WLAN_GETMIB:
			{
				UINT8*	p = NULL;
				UINT32	size = 0;
				UINT32	mibcode;
				p = kmalloc(req->len, GFP_KERNEL);
				if (p == NULL)
				{
					printk(KERN_DEBUG "kmalloc failed in ioctl.\n");
					break;
				}
				copy_from_user( &mibcode, req->data, 4 );/* mibcode frm wlan_mibget_t*/
				switch(mibcode) 
				{
					case SUMIB_LOCAL:
						size = sizeof(su_mib_local_t);
						break;
					case SUMIB_ADDR:
						size = sizeof(su_mib_mac_addr_stat_grp_t);
						break;
					case SUMIB_MAC:
						size = sizeof(su_mib_mac_t); 
						break;
					case SUMIB_STAT:
						size = sizeof(su_mib_mac_statistics_t); 
						break;
					case SUMIB_MGMT:
						size = sizeof(su_mib_mac_mgmt_t); 
						break;
					case SUMIB_DRVR:
						/* size = sizeof(su_mib_local_t); */
						size = 0;
						break;
					case SUMIB_PHY:
						size = sizeof(su_mib_phy_t); 
						break;
				}

				am930mgr_mibget( V2P(dev->priv)->llc->mgr, mibcode, size, p);

				copy_to_user( ((UINT8*)(req->data)) + 4, p, size);
				kfree_s(p, req->len);
				break;
			}
			case WLAN_GETMIBITEM:
				break;
			case WLAN_SETMIBITEM:
				break;
			default:
				printk(KERN_DEBUG "am930_cs: Unknown or unsupported wlan ioctl!\n");
				break;
		}
	}

	DBFEXIT;
	return result;
}

/*----------------------------------------------------------------
*	am930llc_ontxcomplete
*
*	returns: zero
----------------------------------------------------------------*/
void am930llc_ontxcomplete( am930llc_t *llc, UINT32 txresult)
{
	DBFENTER;

	if ( txresult != 0 ) /* tx failure */
	{
		V2P(llc->dev->priv)->stats.tx_errors++;
	}
	else
	{
		V2P(llc->dev->priv)->stats.tx_packets++;
	}

	llc->dev->tbusy = 0;
	mark_bh(NET_BH);

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930llc_dbprt_frame
*
*	returns: nothing
----------------------------------------------------------------*/
void am930llc_dbprt_frame( UINT8* frm, UINT32 len, int isrx )
{
	int i;
	char buf[80];

	if ( len > 0 )
	{
		for ( i = 0; i < len; i++)
		{
			if ( i % 20 == 0 && i > 0)
			{
				printk( KERN_DEBUG "llc: %s\n", buf);
			}
			sprintf( &buf[(i % 20)*3], "%02x ", frm[i]);
		}
	
		if ( i % 20 > 0 )
		{
			printk( KERN_DEBUG "%s%d: %s\n", isrx ? "rx" : "tx", i % 20, buf);
		}
	}
	else
	{
		printk( KERN_DEBUG "llc: frame length is zero\n");
	}
}
